// const { rules: coreStylistic } = require('./core-stylistic-issues');
// const { rules: coreBestPractices } = require('./core-best-practices');
// const { rules: coreVariables } = require('./core-variables');
// const { rules: corePossibleErrors } = require('./core-possible-errors');
// const { rules: coreES6 } = require('./core-ecmascript-6');

// ----------------------------------------------------------------------------
// Typescript
// https://github.com/typescript-eslint/typescript-eslint
// ----------------------------------------------------------------------------

const rules = {
    // Require that member overloads be consecutive
    '@typescript-eslint/adjacent-overload-signatures': 'error',

    // Requires using either T[] or Array<T> for arrays
    '@typescript-eslint/array-type': [
        'error',
        { default: 'array-simple', readonly: 'array-simple' },
    ],

    // Disallows awaiting a value that is not a Thenable
    '@typescript-eslint/await-thenable': 'error', // TODO

    // Bans "// @ts-<directive>" comments from being used
    '@typescript-eslint/ban-ts-comment': 'error',

    // // Bans "// @ts-ignore" comments from being used
    // '@typescript-eslint/ban-ts-ignore': 'error',
    // DEPRECATED ban-ts-comment

    // Bans specific types from being used
    '@typescript-eslint/ban-types': [
        'error',
        {
            types: {
                String: { message: 'Use string instead', fixWith: 'string' },
                Boolean: { message: 'Use boolean instead', fixWith: 'boolean' },
                Number: { message: 'Use number instead', fixWith: 'number' },
                Object: { message: 'Use object instead', fixWith: 'object' },
                // Array: { message: 'Provide a more specific type' },
                Function: {
                    message: 'Provide a more specific type, like `() => void`.',
                },
            },
        },
    ],

    // Ensures that literals on classes are exposed in a consistent style
    '@typescript-eslint/class-literal-property-style': ['error', 'fields'],

    // // Require PascalCased class and interface names
    // '@typescript-eslint/class-name-casing': 'error',
    // DEPRECATED naming-convention

    // Enforces consistent usage of type assertions
    '@typescript-eslint/consistent-type-assertions': [
        'error',
        { assertionStyle: 'as', objectLiteralTypeAssertions: 'allow' },
    ],

    // Consistent with type definition either interface or type
    '@typescript-eslint/consistent-type-definitions': 'off',

    // Require explicit return types on functions and class methods
    '@typescript-eslint/explicit-function-return-type': 'off', // TODO

    // Require explicit accessibility modifiers on class properties and methods
    '@typescript-eslint/explicit-member-accessibility': ['error', { accessibility: 'no-public' }], // TODO

    // Require explicit return and argument types on exported functions' and classes' public class methods
    '@typescript-eslint/explicit-module-boundary-types': 'off', // TODO error

    // // Enforces naming of generic type variables
    // '@typescript-eslint/generic-type-naming': 'off',
    // DEPRECATED naming-convention

    // // Require that interface names should or should not prefixed with I
    // '@typescript-eslint/interface-name-prefix': [
    //     'error',
    //     { prefixWithI: 'always', allowUnderscorePrefix: false },
    // ],
    // DEPRECATED naming-convention

    // Require a specific member delimiter style for interfaces and type literals
    '@typescript-eslint/member-delimiter-style': [
        'error',
        {
            multiline: {
                delimiter: 'semi',
                requireLast: true,
            },
            singleline: {
                delimiter: 'semi',
                requireLast: true,
            },
        },
    ],

    // // Enforces naming conventions for class members by visibility
    // '@typescript-eslint/member-naming': 'off',
    // DEPRECATED naming-convention

    // Require a consistent member declaration order
    '@typescript-eslint/member-ordering': [
        'error',
        {
            default: [
                'public-static-field',
                'protected-static-field',
                'private-static-field',
                'public-static-method',
                'protected-static-method',
                'private-static-method',
                'public-instance-field',
                'protected-instance-field',
                'private-instance-field',
                'constructor',
                'public-instance-method',
                'protected-instance-method',
                'private-instance-method',
            ],
        },
    ],

    // Enforces using a particular method signature syntax.
    '@typescript-eslint/method-signature-style': ['error', 'property'],

    // Enforces naming conventions for everything across a codebase
    '@typescript-eslint/naming-convention': [
        // TODO enable
        'off',
        {
            selector: 'default',
            format: ['camelCase'],
        },
        {
            selector: 'variable',
            format: ['camelCase', 'PascalCase', 'UPPER_CASE'],
        },
        {
            selector: 'parameter',
            format: ['camelCase'],
            leadingUnderscore: 'allow',
        },
        {
            selector: 'enumMember',
            format: ['camelCase', 'UPPER_CASE'],
        },
        {
            selector: 'memberLike',
            modifiers: ['private'],
            format: ['camelCase'],
            leadingUnderscore: 'require',
        },
        {
            selector: 'typeLike',
            format: ['PascalCase'],
        },
    ],

    // Requires that .toString() is only called on objects which provide useful information when stringified
    '@typescript-eslint/no-base-to-string': 'error',

    // Disallow the delete operator with computed key expressions
    '@typescript-eslint/no-dynamic-delete': 'error',

    // Disallow the declaration of empty interfaces
    '@typescript-eslint/no-empty-interface': 'off',

    // Disallow usage of the any type
    '@typescript-eslint/no-explicit-any': 'warn',

    // This rule disallows poor use of null assertion
    '@typescript-eslint/no-extra-non-null-assertion': 'error',

    // Forbids the use of classes as namespaces
    '@typescript-eslint/no-extraneous-class': 'error',

    // Requires Promise-like values to be handled appropriately
    '@typescript-eslint/no-floating-promises': 'error',

    // Disallow iterating over an array with a for-in loop
    '@typescript-eslint/no-for-in-array': 'error',

    // Disallow the use of eval()-like methods
    '@typescript-eslint/no-implied-eval': 'error',

    // Disallows explicit type declarations for variables or parameters initialized to a number, string, or boolean
    '@typescript-eslint/no-inferrable-types': [
        'error',
        {
            ignoreParameters: false,
            ignoreProperties: false,
        },
    ],

    // Disallows usage of void type outside of generic or return types
    '@typescript-eslint/no-invalid-void-type': ['warn', { allowInGenericTypeArguments: true }], // TODO

    // Enforce valid definition of new and constructor
    '@typescript-eslint/no-misused-new': 'error',

    // Avoid using promises in places not designed to handle them
    '@typescript-eslint/no-misused-promises': [
        'error',
        {
            checksVoidReturn: true,
            checksConditionals: true,
        },
    ],

    // Disallow the use of custom TypeScript modules and namespaces
    '@typescript-eslint/no-namespace': [
        'error',
        { allowDeclarations: false, allowDefinitionFiles: false },
    ],

    // Disallows using a non-null assertion after an optional chain expression
    '@typescript-eslint/no-non-null-asserted-optional-chain': 'error',

    // Disallows non-null assertions using the ! postfix operator
    '@typescript-eslint/no-non-null-assertion': 'error', // TODO

    // Disallow the use of parameter properties in class constructors
    '@typescript-eslint/no-parameter-properties': 'off',

    // Disallows invocation of require()
    '@typescript-eslint/no-require-imports': 'off',

    // Disallow aliasing this
    '@typescript-eslint/no-this-alias': ['error', { allowDestructuring: true }],

    // Disallow throwing literals as exceptions
    '@typescript-eslint/no-throw-literal': 'error',

    // Disallow the use of type aliases
    '@typescript-eslint/no-type-alias': 'off',

    // Flags unnecessary equality comparisons against boolean literals
    '@typescript-eslint/no-unnecessary-boolean-literal-compare': 'error',

    // Prevents conditionals where the type is always truthy or always falsy
    '@typescript-eslint/no-unnecessary-condition': [
        'error',
        { ignoreRhs: true, allowConstantLoopConditions: true, checkArrayPredicates: true },
    ], // TODO

    // Warns when a namespace qualifier is unnecessary
    '@typescript-eslint/no-unnecessary-qualifier': 'error',

    // Warns if an explicitly specified type argument is the default for that type parameter
    '@typescript-eslint/no-unnecessary-type-arguments': 'error',

    // Warns if a type assertion does not change the type of an expression
    '@typescript-eslint/no-unnecessary-type-assertion': 'off', // Disabling because in most cases the explicitness is still valuable

    // Disallows assigning any to variables and properties
    '@typescript-eslint/no-unsafe-assignment': 'error',

    // Disallows calling an any type value
    '@typescript-eslint/no-unsafe-call': 'error',

    // Disallows member access on any typed variables
    '@typescript-eslint/no-unsafe-member-access': 'error',

    // Disallows returning any from a function
    '@typescript-eslint/no-unsafe-return': 'error',

    // Disallow unused variables and arguments
    '@typescript-eslint/no-unused-vars-experimental': 'off',

    // // This rule aims to ensure that only typed public methods are declared in the code
    // '@typescript-eslint/no-untyped-public-signature': 'warn',
    // DEPRECATED explicit-module-boundary-types

    // Disallows the use of require statements except in import statements
    '@typescript-eslint/no-var-requires': 'error',

    // Prefer usage of as const over literal type
    '@typescript-eslint/prefer-as-const': 'error',

    // Prefer a ‘for-of’ loop over a standard ‘for’ loop if the index is only used to access the array being iterated
    '@typescript-eslint/prefer-for-of': 'off',

    // Use function types instead of interfaces with call signatures
    '@typescript-eslint/prefer-function-type': 'off',

    // Enforce includes method over indexOf method
    '@typescript-eslint/prefer-includes': 'off',

    // Require the use of the namespace keyword instead of the module keyword to declare custom TypeScript modules
    '@typescript-eslint/prefer-namespace-keyword': 'off',

    // Enforce the usage of the nullish coalescing operator instead of logical chaining
    '@typescript-eslint/prefer-nullish-coalescing': 'error',

    // Prefer using concise optional chain expressions instead of chained logical ands
    '@typescript-eslint/prefer-optional-chain': 'error',

    // Requires that private members are marked as readonly if they're never modified outside of the constructor
    '@typescript-eslint/prefer-readonly': ['error', { onlyInlineLambdas: false }],

    // Requires that function parameters are typed as readonly to prevent accidental mutation of inputs
    '@typescript-eslint/prefer-readonly-parameter-types': 'off', // TODO

    // Prefer using type parameter when calling Array#reduce instead of casting
    '@typescript-eslint/prefer-reduce-type-parameter': 'error',

    // Prefer RegExp#exec() over String#match() if no global flag is provided
    '@typescript-eslint/prefer-regexp-exec': 'error',

    // Enforce the use of String#startsWith and String#endsWith instead of other equivalent methods of checking substrings
    '@typescript-eslint/prefer-string-starts-ends-with': 'off',

    // Recommends using // @ts-expect-error over // @ts-ignore
    '@typescript-eslint/prefer-ts-expect-error': 'error',

    // Requires any function or method that returns a Promise to be marked async
    '@typescript-eslint/promise-function-async': 'off',

    // Requires Array#sort calls to always provide a compareFunction
    '@typescript-eslint/require-array-sort-compare': 'off',

    // When adding two variables, operands must both be of type number or of type string
    '@typescript-eslint/restrict-plus-operands': 'error',

    // Enforce template literal expressions to be of string type
    '@typescript-eslint/restrict-template-expressions': 'off', // TODO

    // Restricts the types allowed in boolean expressions
    '@typescript-eslint/strict-boolean-expressions': 'off',

    // Exhaustiveness checking in switch with union type
    '@typescript-eslint/switch-exhaustiveness-check': 'error',

    // Sets preference level for triple slash directives versus ES6-style import declarations
    '@typescript-eslint/triple-slash-reference': [
        'error',
        {
            path: 'never',
            types: 'prefer-import', // TODO never
            lib: 'never',
        },
    ],

    // Require consistent spacing around type annotations
    '@typescript-eslint/type-annotation-spacing': ['error'],

    // Requires type annotations to exist
    '@typescript-eslint/typedef': 'off',

    // Enforces unbound methods are called with their expected scope
    '@typescript-eslint/unbound-method': 'off',

    // Warns for any two overloads that could be unified into one by using a union or an optional/rest parameter
    '@typescript-eslint/unified-signatures': 'off',

    // ----------------------------------------------------------------------------
    // Eslint overrides
    // ----------------------------------------------------------------------------

    // Enforce consistent brace style for blocks
    'brace-style': 'off',
    // '@typescript-eslint/brace-style': coreStylistic['brace-style'],
    '@typescript-eslint/brace-style': ['error', '1tbs', { allowSingleLine: false }],

    // Enforce camelCase naming convention
    camelcase: 'off',
    'babel/camelcase': 'off',
    // '@typescript-eslint/camelcase': ['error', { properties: 'always' }],
    // DEPRECATED naming-convention

    // Enforces consistent spacing before and after commas
    'comma-spacing': 'off',
    // '@typescript-eslint/comma-spacing': coreStylistic['comma-spacing'],
    '@typescript-eslint/comma-spacing': ['error', { before: false, after: true }],

    // Enforce default parameters to be last
    'default-param-last': 'off',
    // '@typescript-eslint/default-param-last': coreBestPractices['default-param-last'],
    '@typescript-eslint/default-param-last': 'off',

    // enforce dot notation whenever possible
    'dot-notation': 'off',
    // '@typescript-eslint/dot-notation': coreBestPractices['dot-notation'],
    '@typescript-eslint/dot-notation': ['error', { allowKeywords: true }],

    // Require or disallow spacing between function identifiers and their invocations
    'func-call-spacing': 'off',
    // '@typescript-eslint/func-call-spacing': coreStylistic['func-call-spacing'],
    '@typescript-eslint/func-call-spacing': ['error', 'never'],

    // Enforce consistent indentation
    indent: 'off',
    // '@typescript-eslint/indent': coreStylistic['indent'],
    '@typescript-eslint/indent': [
        'error',
        4,
        {
            SwitchCase: 1,
            VariableDeclarator: 'first',
        },
    ],

    // require or disallow initialization in variable declarations
    'init-declarations': 'off',
    // '@typescript-eslint/init-declarations': coreVariables['init-declarations'],
    '@typescript-eslint/init-declarations': 'off',

    // Enforce consistent spacing before and after keywords
    'keyword-spacing': 'off',
    // '@typescript-eslint/keyword-spacing': coreStylistic['keyword-spacing'],
    '@typescript-eslint/keyword-spacing': ['error', { before: true, after: true }],

    // Require or disallow an empty line between class members
    'lines-between-class-members': 'off',
    // '@typescript-eslint/lines-between-class-members': coreStylistic['lines-between-class-members'],
    '@typescript-eslint/lines-between-class-members': [
        'error',
        'always',
        { exceptAfterSingleLine: false },
    ],

    // Disallow generic Array constructors
    'no-array-constructor': 'off',
    // '@typescript-eslint/no-array-constructor': coreStylistic['no-array-constructor'],
    '@typescript-eslint/no-array-constructor': 'error',

    // Disallow duplicate class members
    'no-dupe-class-members': 'off',
    // '@typescript-eslint/no-dupe-class-members': coreES6['no-dupe-class-members'],
    '@typescript-eslint/no-dupe-class-members': 'error',

    // Disallow empty functions
    'no-empty-function': 'off',
    // '@typescript-eslint/no-empty-function': coreBestPractices['no-empty-function'],
    '@typescript-eslint/no-empty-function': 'error',

    // Disallow unnecessary parentheses
    'no-extra-parens': 'off',
    // '@typescript-eslint/no-extra-parens': corePossibleErrors['no-extra-parens'],
    '@typescript-eslint/no-extra-parens': [
        // TODO
        'warn',
        'all',
        {
            conditionalAssign: true,
            returnAssign: false,
            nestedBinaryExpressions: false,
            ignoreJSX: 'multi-line',
            enforceForArrowConditionals: false,
            enforceForSequenceExpressions: true,
            enforceForNewInMemberExpressions: false,
        },
    ],

    // Disallow unnecessary semicolons
    'no-extra-semi': 'off',
    // '@typescript-eslint/no-extra-semi': corePossibleErrors['no-extra-semi'],
    '@typescript-eslint/no-extra-semi': 'error',

    // disallow this keywords outside of classes or class-like objects
    'no-invalid-this': 'off',
    'babel/no-invalid-this': 'off',
    // '@typescript-eslint/no-invalid-this': coreBestPractices['no-invalid-this'],
    '@typescript-eslint/no-invalid-this': 'off',

    // Disallows magic numbers
    'no-magic-numbers': 'off',
    // '@typescript-eslint/no-magic-numbers': coreBestPractices['no-magic-numbers'],
    '@typescript-eslint/no-magic-numbers': 'off',

    // Disallow unused expressions
    'no-unused-expressions': 'off',
    'babel/no-unused-expressions': 'off',
    // '@typescript-eslint/no-unused-expressions': coreBestPractices['no-unused-expressions'],
    '@typescript-eslint/no-unused-expressions': [
        // TODO
        'error',
        {
            allowShortCircuit: false,
            allowTernary: false,
            allowTaggedTemplates: false,
        },
    ],

    // Disallow unused variables
    'no-unused-vars': 'off',
    '@typescript-eslint/no-unused-vars': 'off',

    // Disallow the use of variables before they are defined
    'no-use-before-define': 'off',
    '@typescript-eslint/no-use-before-define': 'off', // TODO

    // Disallow unnecessary constructors
    'no-useless-constructor': 'off',
    // '@typescript-eslint/no-useless-constructor': coreES6['no-useless-constructor'],
    '@typescript-eslint/no-useless-constructor': 'error',

    // Enforce the consistent use of either backticks, double, or single quotes
    quotes: 'off',
    'babel/quotes': 'off',
    // '@typescript-eslint/quotes': coreStylistic['quotes'],
    '@typescript-eslint/quotes': [
        // TODO
        'error',
        'single',
        {
            avoidEscape: true,
            allowTemplateLiterals: true,
        },
    ],

    // Disallow async functions which have no await expression
    'require-await': 'off',
    // '@typescript-eslint/require-await': coreBestPractices['require-await'],
    '@typescript-eslint/require-await': 'error',

    // Enforces consistent returning of awaited values
    'no-return-await': 'off',
    // '@typescript-eslint/return-await': coreBestPractices['no-return-await'],
    '@typescript-eslint/return-await': 'error',

    // Require or disallow semicolons instead of ASI
    semi: 'off',
    'babel/semi': 'off',
    // '@typescript-eslint/semi': coreStylistic['semi'],
    '@typescript-eslint/semi': ['error', 'always'],

    // Enforces consistent spacing before function parenthesis
    'space-before-function-paren': 'off',
    // '@typescript-eslint/space-before-function-paren': coreStylistic['space-before-function-paren'],
    '@typescript-eslint/space-before-function-paren': [
        // TODO
        'error',
        {
            anonymous: 'always', // TODO never?
            named: 'never',
            asyncArrow: 'always',
        },
    ],
};

module.exports = {
    plugins: ['@typescript-eslint'],

    overrides: [
        {
            files: ['*.ts', '*.tsx'],
            rules,
        },
        {
            files: ['*.*{spec,test}.{ts,tsx}', '**/__{mocks,tests}__/**/*.{ts,tsx}'],
            rules: {
                '@typescript-eslint/no-unsafe-assignment': 'warn',
                '@typescript-eslint/no-unsafe-call': 'warn',
                '@typescript-eslint/no-unsafe-member-access': 'warn',
                '@typescript-eslint/no-unsafe-return': 'warn',
            },
        },
    ],
};
